@@ -1,19 +1,21 @@
-"""
-bundlerepo.py - repository class for viewing uncompressed bundles
+# bundlerepo.py - repository class for viewing uncompressed bundles
+#
+# Copyright 2006, 2007 Benoit Boissinot <bboissin@gmail.com>
+#
+# This software may be used and distributed according to the terms of the
+# GNU General Public License version 2, incorporated herein by reference.
 
-This provides a read-only repository interface to bundles as if
-they were part of the actual repository.
+"""Repository class for viewing uncompressed bundles.
 
-Copyright 2006, 2007 Benoit Boissinot <bboissin@gmail.com>
-
-This software may be used and distributed according to the terms
-of the GNU General Public License, incorporated herein by reference.
+This provides a read-only repository interface to bundles as if they
+were part of the actual repository.
 """
 
-from node import hex, nullid, short
+from node import nullid
 from i18n import _
-import changegroup, util, os, struct, bz2, zlib, tempfile, shutil, mdiff
-import repo, localrepo, changelog, manifest, filelog, revlog, context
+import os, struct, bz2, zlib, tempfile, shutil
+import changegroup, util, mdiff
+import localrepo, changelog, manifest, filelog, revlog, error
 
 class bundlerevlog(revlog.revlog):
     def __init__(self, opener, indexfile, bundlefile,
@@ -48,8 +50,8 @@ class bundlerevlog(revlog.revlog):
                 continue
             for p in (p1, p2):
                 if not p in self.nodemap:
-                    raise revlog.LookupError(p1, self.indexfile,
-                                             _("unknown parent"))
+                    raise error.LookupError(p, self.indexfile,
+                                            _("unknown parent"))
             if linkmapper is None:
                 link = n
             else:
@@ -72,12 +74,12 @@ class bundlerevlog(revlog.revlog):
             return False
         return rev in self.basemap
     def bundlebase(self, rev): return self.basemap[rev]
-    def chunk(self, rev, df=None, cachelen=4096):
+    def _chunk(self, rev):
         # Warning: in case of bundle, the diff is against bundlebase,
         # not against rev - 1
         # XXX: could use some caching
         if not self.bundle(rev):
-            return revlog.revlog.chunk(self, rev, df)
+            return revlog.revlog._chunk(self, rev)
         self.bundlefile.seek(self.start(rev))
         return self.bundlefile.read(self.length(rev))
 
@@ -87,7 +89,7 @@ class bundlerevlog(revlog.revlog):
             # hot path for bundle
             revb = self.rev(self.bundlebase(rev2))
             if revb == rev1:
-                return self.chunk(rev2)
+                return self._chunk(rev2)
         elif not self.bundle(rev1) and not self.bundle(rev2):
             return revlog.revlog.revdiff(self, rev1, rev2)
 
@@ -114,12 +116,12 @@ class bundlerevlog(revlog.revlog):
             text = revlog.revlog.revision(self, iter_node)
 
         while chain:
-            delta = self.chunk(chain.pop())
+            delta = self._chunk(chain.pop())
             text = mdiff.patches(text, [delta])
 
         p1, p2 = self.parents(node)
         if node != revlog.hash(text, p1, p2):
-            raise revlog.RevlogError(_("integrity check failed on %s:%d")
+            raise error.RevlogError(_("integrity check failed on %s:%d")
                                      % (self.datafile, self.rev(node)))
 
         self._cache = (node, self.rev(node), text)
@@ -156,9 +158,9 @@ class bundlerepository(localrepo.localrepository):
         self._tempparent = None
         try:
             localrepo.localrepository.__init__(self, ui, path)
-        except repo.RepoError:
+        except error.RepoError:
             self._tempparent = tempfile.mkdtemp()
-            tmprepo = localrepo.instance(ui,self._tempparent,1)
+            localrepo.instance(ui, self._tempparent, 1)
             localrepo.localrepository.__init__(self, ui, self._tempparent)
 
         if path:
@@ -208,25 +210,28 @@ class bundlerepository(localrepo.localrepository):
         # dict with the mapping 'filename' -> position in the bundle
         self.bundlefilespos = {}
 
-    def __getattr__(self, name):
-        if name == 'changelog':
-            self.changelog = bundlechangelog(self.sopener, self.bundlefile)
-            self.manstart = self.bundlefile.tell()
-            return self.changelog
-        elif name == 'manifest':
-            self.bundlefile.seek(self.manstart)
-            self.manifest = bundlemanifest(self.sopener, self.bundlefile,
-                                           self.changelog.rev)
-            self.filestart = self.bundlefile.tell()
-            return self.manifest
-        elif name == 'manstart':
-            self.changelog
-            return self.manstart
-        elif name == 'filestart':
-            self.manifest
-            return self.filestart
-        else:
-            raise AttributeError(name)
+    @util.propertycache
+    def changelog(self):
+        c = bundlechangelog(self.sopener, self.bundlefile)
+        self.manstart = self.bundlefile.tell()
+        return c
+
+    @util.propertycache
+    def manifest(self):
+        self.bundlefile.seek(self.manstart)
+        m = bundlemanifest(self.sopener, self.bundlefile, self.changelog.rev)
+        self.filestart = self.bundlefile.tell()
+        return m
+
+    @util.propertycache
+    def manstart(self):
+        self.changelog
+        return self.manstart
+
+    @util.propertycache
+    def filestart(self):
+        self.manifest
+        return self.filestart
 
     def url(self):
         return self._url
